// include external files and initialise MIDI library
#include "AudioConnections.h"
#include "KeyboardHandler.h"
#include <MIDI.h>
MIDI_CREATE_INSTANCE(HardwareSerial, Serial1, MIDI);

// array of references to input patch points (2 arrays for 2-note polyphony)
AudioMixer4 *inputMixers[2][8][3] = {
  {
    {&sawtoothModMixerA1,&sawtoothModMixerB1,&sawtoothModMixerC1},
    {&squareModMixerA1,&squareModMixerB1,&squareModMixerC1},
    {&filterInputMixerA1,&filterInputMixerB1,&filterInputMixerC1},
    {&filterModMixerA1,&filterModMixerB1,&filterModMixerC1},
    {&ampMixerA1,&ampMixerB1,&ampMixerC1},
    {&ampControlMixerA1,&ampControlMixerB1,&ampControlMixerC1},
    {&envelopeMixerA1,&envelopeMixerB1,&envelopeMixerC1},
    {&mainMixerA1,&mainMixerB1,&mainMixerC1}
  },{
    {&sawtoothModMixerA2,&sawtoothModMixerB2,&sawtoothModMixerC2},
    {&squareModMixerA2,&squareModMixerB2,&squareModMixerC2},
    {&filterInputMixerA2,&filterInputMixerB2,&filterInputMixerC2},
    {&filterModMixerA2,&filterModMixerB2,&filterModMixerC2},
    {&ampMixerA2,&ampMixerB2,&ampMixerC2},
    {&ampControlMixerA2,&ampControlMixerB2,&ampControlMixerC2},
    {&envelopeMixerA2,&envelopeMixerB2,&envelopeMixerC2},
    {&mainMixerA2,&mainMixerB2,&mainMixerC2}
  }
};

// define pins
const int ADDRESS_SEND_PINS[] = {2,3,4};
const int ADDRESS_READ_PINS[] = {5,6,7};
const int CONNECTION_SEND_PINS[] = {12, 10};
const int CONNECTION_READ_PINS[] = {17, 8};
const int KEYBOARD_PIN = 16;
const int ANALOG_PIN_1 = 20;
const int ANALOG_PIN_2 = 21;
const int LED_PIN = 14;

// global variables
boolean patchConnections[16][16];
KeyboardHandler keyboardHandler;
const int buttonNotes[] = {36, 38, 40, 41, 43, 45, 47, 48};
boolean useBreadboardKeyboard = true; // this will remain true until a MIDI signal is detected, when the breadboard keyboard will be disabled

void setup() {
  // initialise pins
  for(int i=0;i<3;i++) {
    pinMode(ADDRESS_SEND_PINS[i], OUTPUT);
    pinMode(ADDRESS_READ_PINS[i], OUTPUT);
  }
  pinMode(CONNECTION_READ_PINS[0], INPUT_PULLUP);
  pinMode(CONNECTION_READ_PINS[1], INPUT_PULLUP);
  pinMode(KEYBOARD_PIN, INPUT_PULLUP);
  pinMode(LED_PIN, OUTPUT);

  // initialise audio library
  AudioMemory(100);
  sgtl5000_1.enable();
  sgtl5000_1.volume(0.4);

  Serial.begin(9600);

  MIDI.begin(MIDI_CHANNEL_OMNI);
  
  // initialise audio objects - repeated twice for 2-note polyphony
  square1.begin(0.1, 440, WAVEFORM_SQUARE);
  sawtooth1.begin(0.1, 440, WAVEFORM_SAWTOOTH);
  lfo1.begin(0.5, 4, WAVEFORM_SINE);
  filter1.octaveControl(4);
  filter1.resonance(2.5);

  envelopeGateInput1.begin();
  envelope1.attack(2);
  envelope1.sustain(0.2);
  envelope1.release(500);
  dc1.amplitude(1.0);
  keyboardCV1.amplitude(0.0);

  square2.begin(0.1, 440, WAVEFORM_SQUARE);
  sawtooth2.begin(0.1, 440, WAVEFORM_SAWTOOTH);
  filter2.octaveControl(4);
  filter2.resonance(2.5);

  envelopeGateInput2.begin();
  envelope2.attack(2);
  envelope2.sustain(0.2);
  envelope2.release(500);
  dc2.amplitude(1.0);
  keyboardCV2.amplitude(0.0);
  
  polyMixer.gain(0,1.0);
  polyMixer.gain(1,1.0);
}

void loop() {
  boolean connectionReading;
  boolean anyBadConnections = false; // keep track of bad connections so we can light LED if needed

  // multi-level for-loop iterates over both sets of 4051 chips
  for(int a=0;a<2;a++) {
    if(a==0) {
      pinMode(CONNECTION_SEND_PINS[0], OUTPUT);
      pinMode(CONNECTION_SEND_PINS[1], INPUT);
    } else {
      pinMode(CONNECTION_SEND_PINS[1], OUTPUT);
      pinMode(CONNECTION_SEND_PINS[0], INPUT);
    }
    digitalWrite(CONNECTION_SEND_PINS[a], LOW);
    
    for(int b=0;b<8;b++) {
      setSendChannel(b);

      for(int c=0;c<2;c++) {
  
        for(int d=0;d<8;d++) {
          checkMidi();
          
          setReadChannel(d);
    
          delayMicroseconds(10);
          int pin1 = b + a*8;
          int pin2 = d + c*8;
          if(pin1 < pin2) {
            connectionReading = !digitalRead(CONNECTION_READ_PINS[c]);
            if(connectionReading != patchConnections[pin1][pin2]) {
              patchConnections[pin1][pin2] = connectionReading;
              if(connectionReading) {
                Serial.print(pin1);
                Serial.print(" connected to ");
                Serial.println(pin2);
              } else {
                Serial.print(pin1);
                Serial.print(" disconnected from ");
                Serial.println(pin2);
              }
            }
            if(connectionReading) {
              if(connectionReading && pin1 <= 7 && pin2 >= 8) {
                // new connection valid (output to input)
                inputMixers[0][pin2-8][pin1/4]->gain(pin1%4,1.0);
                inputMixers[1][pin2-8][pin1/4]->gain(pin1%4,1.0);
              } else {
                // new connection not valid (output to output or input to input)
                anyBadConnections = true;
              }
            } else {
              if(pin1 <= 7 && pin2 >= 8) { 
                inputMixers[0][pin2-8][pin1/4]->gain(pin1%4,0.0);
                inputMixers[1][pin2-8][pin1/4]->gain(pin1%4,0.0);
              }
            }
          }
          if(useBreadboardKeyboard) {
            keyboardHandler.setKey(buttonNotes[d], !digitalRead(KEYBOARD_PIN));
            keyboardHandler.update();
          }
          checkEnvelopeGates();
          filter1.frequency(4*analogRead(ANALOG_PIN_1));
          filter2.frequency(4*analogRead(ANALOG_PIN_1));
          lfo1.frequency(analogRead(ANALOG_PIN_2)/16);
        }
        // convert MIDI note number to virtual "control voltage"
        // the 12.0*8 term comes from 12 notes per octave and 8 octaves per "volt" in the virtual signal
        keyboardCV1.amplitude((keyboardHandler.getNote(0)-60)/(12.0*8));
        keyboardCV2.amplitude((keyboardHandler.getNote(1)-60)/(12.0*8));
        
        keyboardGate1.amplitude(keyboardHandler.getGate(0)?1.0:0.0);
        keyboardGate2.amplitude(keyboardHandler.getGate(1)?1.0:0.0);
      }
    }
  }
  // light LED if any bad connections
  digitalWrite(LED_PIN, anyBadConnections);
}

void setSendChannel(int channel) {
  digitalWrite(ADDRESS_SEND_PINS[0], bitRead(channel, 0));
  digitalWrite(ADDRESS_SEND_PINS[1], bitRead(channel, 1));
  digitalWrite(ADDRESS_SEND_PINS[2], bitRead(channel, 2));
}

void setReadChannel(int channel) {
  digitalWrite(ADDRESS_READ_PINS[0], bitRead(channel, 0));
  digitalWrite(ADDRESS_READ_PINS[1], bitRead(channel, 1));
  digitalWrite(ADDRESS_READ_PINS[2], bitRead(channel, 2));
}

// handle incoming MIDI messages
void checkMidi() {
  int note, velocity;
  if (MIDI.read()) {
    useBreadboardKeyboard = false;
    byte type = MIDI.getType();
    switch (type) {
      case midi::NoteOn:
        note = MIDI.getData1();
        velocity = MIDI.getData2();
        if (velocity > 0) {
          keyboardHandler.setKey(note, true);
          keyboardHandler.update();
        } else {
          // sometimes a MIDI "note on" message with zero velocity is send instead of a "note off" message
          keyboardHandler.setKey(note, false);
          keyboardHandler.update();
        }
        break;
      case midi::NoteOff:
        note = MIDI.getData1();
        keyboardHandler.setKey(note, false);
        keyboardHandler.update();
        break;
    }
  }
}

// this section of code allows the envelope to be triggered by a gate (on/off) signal
boolean envelopeGateOpen1 = false;
boolean envelopeGateOpen2 = false;
void checkEnvelopeGates() {
  if(envelopeGateInput1.available() > 0) {
    int16_t test[128];
    memcpy(test, envelopeGateInput1.readBuffer(), 256);
    envelopeGateInput1.freeBuffer();
    int16_t x = test[0];
    boolean newGateReading = (x > 0);
    if(newGateReading != envelopeGateOpen1) {
      envelopeGateOpen1 = newGateReading;
      if(envelopeGateOpen1) {
        envelope1.noteOn();
      }
      else envelope1.noteOff();
    }
    envelopeGateInput1.clear();
  }
  if(envelopeGateInput2.available() > 0) {
    int16_t test[128];
    memcpy(test, envelopeGateInput2.readBuffer(), 256);
    envelopeGateInput2.freeBuffer();
    int16_t x = test[0];
    boolean newGateReading = (x > 0);
    if(newGateReading != envelopeGateOpen2) {
      envelopeGateOpen2 = newGateReading;
      if(envelopeGateOpen2) {
        envelope2.noteOn();
      }
      else envelope2.noteOff();
    }
    envelopeGateInput2.clear();
  }
}

